/*-!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!
! Filename: hex2ram.c      Function: convert Intel HEX file into DIY-Calc  !
! Part of :                   RAM file format                              !
! Module  :                                                                !
!                                                                          !
!                          run tasm with following parameters:             !
!                             tasm -01 -o10 -c -f0 file.asm file.hex       !
!                                                                          !
!                                                                          !
!   Vers     Date    Auth  Changes                                         !
! ------------------------------------------------------------------------ !
!   1.0    12.Jul.06  JH   New.                                            !
!   1.1    20.Jul.06  JH   remove stdin/stdout redirection                 !
!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!-*/


#define VER "1.1"


#ifdef WIN32
#include <windows.h>
#endif

#ifdef __MSDOS__
#include <dir.h>             /* MAXPATH */
#endif

#ifdef __linux__
#include <limits.h>          /* PATH_MAX */
#endif

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>


/******************************  Constants  *******************************/
#define OK        0
#define BUFLEN  200

#ifdef __linux__
#define PROGNAME "hex2ram"
#define MAX_PATH PATH_MAX
#elif defined WIN32
#define PROGNAME "hex2ram.exe"
#elif defined __MSDOS__
#define PROGNAME "HEX2RAM.EXE"
#define MAX_PATH MAXPATH
#else
#error "unsupported platform"
#endif


/*******************************  Macros  *********************************/
#ifndef HIBYTE
#define HIBYTE(x) (((x)>>8)&0xff)
#endif
#ifndef LOBYTE
#define LOBYTE(x) ((x)&0xff)
#endif


/**************************  Function Prototypes  *************************/
int atohex (char *s);             /* convert string to HEX number         */


/*******************************  Globals  ********************************/
enum REC_TYPE
{
  REC_DATA,
  REC_END
};


/* main */
/**************************************************************************/
int main (int argc, char *argv[])
/***************************************************************************
* globals  :                                                               *
* called by:                                                               *
* calls    :                                                               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
***************************************************************************/
{
  char hexname[MAX_PATH], ramname[MAX_PATH], asmname[MAX_PATH];
  FILE *hex, *ram;
  char buf[BUFLEN], tmp[12];
  char *p;
  unsigned int linecnt, bytecount, address, type, i, j;
  unsigned char data[BUFLEN/2], sum, checksum;
  unsigned int startaddr, endaddr;
  time_t now;


  if (argc == 1)
  {             
    fprintf (stderr, "usage: %s hexfile\n", PROGNAME );
    return 0;
  }
  printf ("\n%s V%s, %s\n\n", PROGNAME, VER, __DATE__);
  /* assemble file names */
  strcpy (hexname, argv[1]);
  p = strchr (hexname, '.');
  if (p == NULL)
    strcat (hexname, ".hex");

  strcpy (ramname, argv[1]);
  p = strrchr (ramname, '.');
  if (p)
    *p = '\0';
  strcat (ramname, ".ram");

  strcpy (asmname, argv[1]);
  p = strrchr (asmname, '.');
  if (p)
    *p = '\0';
  strcat (asmname, ".asm");

  /* open the files */
  hex = fopen (hexname, "rt");
  if (hex == NULL)
  {
    perror (hexname);
    return 1;
  }

  ram = fopen (ramname, "wt");
  if (ram == NULL)
  {
    fclose (hex);
    perror (ramname);
    return 1;
  }

  printf ("\nProcessing file \"%s\" ...    ", hexname);

  /* we now have opened the files; lets do stuff */
  linecnt = 0;

  /* write the .ram file header */
  now = time (NULL);
  fprintf (ram, "# FILE TYPE: DIY Calculator List (*.ram) file\n");
  fprintf (ram, "# GENERATED: %s %s\n", PROGNAME, VER);
  fprintf (ram, "# DATE-TIME: %s", ctime(&now));
  fprintf (ram, "# SOURCEWAS: %s\n\n", asmname);

  /* find the start address end the end address */
  while (fgets (buf, BUFLEN, hex))
  {
    linecnt ++;
    if (buf[0] != ':')
    {
      fprintf (stderr, "\nError line %d: no Intel HEX record\n", linecnt);
      continue;
    }
    tmp[0] = buf[3];
    tmp[1] = buf[4];
    tmp[2] = buf[5];
    tmp[3] = buf[6];
    tmp[4] = '\0';
    address = atohex (tmp);
    if (linecnt == 1)
    {
      startaddr = address;
    }

    tmp[0] = buf[7];
    tmp[1] = buf[8];
    tmp[2] = '\0';
    type = atohex (tmp);

    if (type != REC_END)
    {
      tmp[0] = buf[1];
      tmp[1] = buf[2];
      tmp[2] = '\0';
      bytecount = atohex (tmp);
      endaddr = address + bytecount;
    }
  }
  endaddr --;

  fprintf (ram, "STARTADDR: $%4.4X\n", startaddr);
  fprintf (ram, "FINALADDR: $%4.4X\n\n", endaddr);

  rewind (hex);

  linecnt = 0;
  while (fgets (buf, BUFLEN, hex))
  {
    linecnt ++;
    if (argc > 1)
      printf ("\b\b\b\b%4d", linecnt);

    /* read the .hex file */
    tmp[0] = buf[1];
    tmp[1] = buf[2];
    tmp[2] = '\0';
    bytecount = atohex (tmp);

    tmp[0] = buf[3];
    tmp[1] = buf[4];
    tmp[2] = buf[5];
    tmp[3] = buf[6];
    tmp[4] = '\0';
    address = atohex (tmp);

    tmp[0] = buf[7];
    tmp[1] = buf[8];
    tmp[2] = '\0';
    type = atohex (tmp);

    sum = bytecount + HIBYTE(address) + LOBYTE(address) + type;
    for (i=0; i<=bytecount; i++)
    {
      j = i * 2 + 9;
      tmp[0] = buf[j];
      j ++;
      tmp[1] = buf[j];
      tmp[2] = '\0';
      if (i < bytecount)
      {
        data[i] = atohex (tmp);
        sum += data[i];
      } else
      {
        checksum = atohex (tmp);
        sum = (-sum) & 0xff;
      }
    }
    if (sum != checksum)
    {
      fprintf(stderr,"\nError in line %d: checksum failure (%2.2x/%2.2x)\n",
        linecnt, (unsigned char)sum, (unsigned char)checksum);
    }
    /* write the .ram file */
    if (type != REC_END)
    {
      fprintf (ram, "$%4.4X", address);
      for (i=0; i<bytecount; i++)
        fprintf (ram, " %2.2X", data[i]);
      /* fill last line with dashes */
      for ( ; i<16; i++)
        fprintf (ram, " --");
      fputc ('\n', ram);
    }
  }

  if (argc > 1)
  {
    printf ("\b\b\b\b%4d lines done.\n", linecnt);
    fclose (hex);
    fclose (ram);
  }
  return 0;
}

/* atohex */
/**************************************************************************/
int atohex (char *s)              /* convert string to HEX number         */
/***************************************************************************
* globals     :                                                            *
* called from : main()                                                     *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Convert string to HEX number                                             *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* Zahl (Integer)                                                           *
***************************************************************************/
{
   int i, digit, number;

   i = 0;
   number = 0;
   while (s[i]) {
     switch (toupper(s[i])) {
       case '0': digit =  0; break;
       case '1': digit =  1; break;
       case '2': digit =  2; break;
       case '3': digit =  3; break;
       case '4': digit =  4; break;
       case '5': digit =  5; break;
       case '6': digit =  6; break;
       case '7': digit =  7; break;
       case '8': digit =  8; break;
       case '9': digit =  9; break;
       case 'A': digit = 10; break;
       case 'B': digit = 11; break;
       case 'C': digit = 12; break;
       case 'D': digit = 13; break;
       case 'E': digit = 14; break;
       case 'F': digit = 15; break;
     }
     number = (number << 4) + digit;
     i++;
   }
/* printf("Number = %i, 0x%X\n", number, number);*/
   return (number);
}
